import Middleware from '@poppinss/middleware';
import { moduleImporter } from '@adonisjs/fold';
import { Qs } from '../qs.js';
import debug from '../debug.js';
import { Request } from '../request.js';
import { Response } from '../response.js';
import { Router } from '../router/main.js';
import { HttpContext } from '../http_context/main.js';
import { finalHandler } from './factories/final_handler.js';
import { writeResponse } from './factories/write_response.js';
import { useReturnValue } from './factories/use_return_value.js';
import { asyncLocalStorage } from '../http_context/local_storage.js';
import { middlewareHandler } from './factories/middleware_handler.js';
export class Server {
    #requestHooks = new Set();
    #errorHandler;
    #resolvedErrorHandler = {
        handle(error, ctx) {
            ctx.response.status(error.status || 500).send(error.message || 'Internal server error');
        },
    };
    #app;
    #encryption;
    #config;
    #qsParser;
    #serverMiddlewareStack;
    #router;
    #nodeHttpServer;
    #middleware = [];
    get usingAsyncLocalStorage() {
        return asyncLocalStorage.isEnabled;
    }
    constructor(app, encryption, config) {
        this.#app = app;
        this.#config = config;
        this.#encryption = encryption;
        this.#qsParser = new Qs(this.#config.qs);
        this.#router = new Router(this.#app, this.#encryption, this.#qsParser);
        this.#createAsyncLocalStore();
        debug('server config: %O', this.#config);
    }
    #createAsyncLocalStore() {
        if (this.#config.useAsyncLocalStorage) {
            debug('creating ALS store for HTTP context');
            asyncLocalStorage.create();
        }
        else {
            asyncLocalStorage.destroy();
        }
    }
    #createServerMiddlewareStack() {
        this.#serverMiddlewareStack = new Middleware();
        this.#middleware.forEach((middleware) => this.#serverMiddlewareStack.add(middleware));
        this.#serverMiddlewareStack.freeze();
        this.#middleware = [];
    }
    #handleRequest(ctx, resolver) {
        return this.#serverMiddlewareStack.runner()
            .finalHandler(finalHandler(this.#router, resolver, ctx))
            .run(middlewareHandler(resolver, ctx))
            .then(useReturnValue(ctx))
            .catch((error) => this.#resolvedErrorHandler.handle(error, ctx))
            .finally(writeResponse(ctx));
    }
    use(middleware) {
        middleware.forEach((one) => this.#middleware.push(moduleImporter(one, 'handle').toHandleMethod()));
        return this;
    }
    errorHandler(handler) {
        this.#errorHandler = handler;
        return this;
    }
    async boot() {
        debug('booting HTTP server');
        this.#createServerMiddlewareStack();
        this.#router.commit();
        if (this.#errorHandler) {
            if (debug.enabled) {
                debug('using custom error handler "%s"', this.#errorHandler);
            }
            const moduleExports = await this.#errorHandler();
            this.#resolvedErrorHandler = await this.#app.container.make(moduleExports.default);
        }
    }
    setNodeServer(server) {
        this.#nodeHttpServer = server;
    }
    getNodeServer() {
        return this.#nodeHttpServer;
    }
    getRouter() {
        return this.#router;
    }
    onRequest(callback) {
        this.#requestHooks.add(callback);
        return this;
    }
    handle(req, res) {
        const resolver = this.#app.container.createResolver();
        const request = new Request(req, res, this.#encryption, this.#config, this.#qsParser);
        const response = new Response(req, res, this.#encryption, this.#config, this.#router, this.#qsParser);
        const ctx = new HttpContext(request, response, this.#app.logger.child({}), resolver);
        for (let hook of this.#requestHooks) {
            hook(ctx);
        }
        if (this.usingAsyncLocalStorage) {
            return asyncLocalStorage.storage.run(ctx, () => this.#handleRequest(ctx, resolver));
        }
        return this.#handleRequest(ctx, resolver);
    }
}
